#include <stdio.h>
//#include <string.h>
#include <stdlib.h>
#include <assert.h>

#include <MacTypes.h>
#include <Dialogs.h>
#include <Script.h>
#include <Fonts.h>
#include <Files.h>
#include <TextEdit.h>
#include <Windows.h>
#include <StandardFile.h>
#include <Resources.h>
#include <CodeFragments.h>

#include "mstring.h"
#include "mdebug.h"


typedef short refnum_t;

static void errexit(const char *msg, OSErr err);

class macfile_t {
	private:
		refnum_t	fRefNum;
		bool		fIsOpen;
	
	public:
		macfile_t() : fRefNum(0), fIsOpen(false)
			{ }
		
		~macfile_t()
			{ if (this->fIsOpen) FSClose(this->fRefNum); }
		
		void open_df(const FSSpec *fss, char perm, const char *msg)
		{
			OSErr	err;
			
			assert(!fIsOpen);
			if ((err = FSpOpenDF(fss, perm, &fRefNum)) != noErr)
				errexit(msg, err);
			fIsOpen = true;
		}
		
		void open_rf(const FSSpec *fss, char perm, const char *msg)
		{
			OSErr	err;
			
			assert(!fIsOpen);
			if ((err = FSpOpenRF(fss, perm, &fRefNum)) != noErr)
				errexit(msg, err);
			fIsOpen = true;
		}
		
		void write(SInt32 data, const char *msg)
		{
			OSErr	err;
			long	count = 4;
			
			assert(fIsOpen);
			if ((err = FSWrite(fRefNum, &count, &data)) != noErr)
				errexit(msg, err);
			
			if (count != 4)
				errexit(msg, dskFulErr);
		}
		
		void close()
		{
			assert(fIsOpen);
			FSClose(fRefNum);
			fIsOpen = false;
		}
		
		refnum_t refnum()
		{
			return this->fRefNum;
		}
};

WindowPtr mywindow;

static macfile_t gref, pref, oref, cref;

static void errexit(const char *msg, OSErr err)
{
	Rect	r;
	Str255	buf;
	
	if (err)
		psprintf(buf, "Error: %s (%d)", msg, err);
	else
		StringCopy_pc(buf, msg, sizeof(Str255));

	if (mywindow) {
		HideWindow(mywindow);
		mywindow = 0;
	}
	SetRect(&r, 152, 60, 356, 132);
	SetPort(NewWindow(0L, &r, "\p", 1, 1, (WindowPtr) -1L, 0, 0L));
	TextFont(0);
	MoveTo(4, 20);
	DrawString(buf);
	MoveTo(4, 40);
	DrawString("\pClick mouse to exit.");
	while (!Button())
		;
	
	exit(0);
}

static void filemsg(ConstStr255Param msg)
{
	Rect r;
	
	if (!mywindow) {
		SetRect(&r, 82, 30, 430, 70);
		mywindow = NewWindow(0L, &r, "\p", 1, 1, (WindowPtr) -1L, 0, 0L);
	}
	SetPort(mywindow);
	SetRect(&r, 0, 0, 350, 40);
	EraseRect(&r);
	TextFont(0);
	MoveTo(4, 20);
	DrawString(msg);
}

static void nofilemsg()
{
	HideWindow(mywindow);
	mywindow = 0;
}

static int os_askfile(const char *prompt, FSSpec *file, OSType type, bool writing)
{
	SFTypeList		myTypes;
	SFReply			reply;
	Point			sfgWhere = { 90, 82 };
	Point			sfpWhere = { 106, 104 };
	Str255			pascalprompt;
	static Str255	lastsave = "\pUntitled";

	StringCopy_pc(pascalprompt, prompt, sizeof(pascalprompt));
	myTypes[0] = type;

	if (writing)
		SFPutFile(sfpWhere, pascalprompt, lastsave, 0L, &reply);
	else
		SFGetFile(sfgWhere, pascalprompt, 0L, 1, myTypes, 0L, &reply);

	if (reply.good) {
		FSMakeFSSpec(reply.vRefNum, 0, reply.fName, file);
		return 0;
	}
	else
		return 1;
}

static void copy_fork(refnum_t from, refnum_t to, const char *msg)
{
	static char	buf[16*1024];
	long		count, wcount;
	OSErr		err;
	char		msgbuf[256];
	
	for (;;) {
		count = sizeof(buf);
		err = FSRead(from, &count, buf);
		if (err != noErr && err != eofErr) {
			sprintf(msgbuf, "reading %s", msg);
			errexit(msgbuf, err);
		}
		wcount = count;
		if (!wcount)
			break;
		
		if ((err = FSWrite(to, &wcount, buf)) != noErr) {
			sprintf(msgbuf, "writing %s", msg);
			errexit(msgbuf, err);
		}
			
		if (wcount != count)
			errexit("writing: size mismatch", dskFulErr);
	}
}

inline void copy_fork(macfile_t &from, macfile_t &to, const char *msg)
{
	copy_fork(from.refnum(), to.refnum(), msg);
}

void main()
{
	FSSpec		gamename, progname, origname, cfgname;
	int			vRef = 0;
	OSErr		err;
	FInfo		myinfo;
	DialogPtr	mydlg;
	short		itemHit;
	short		itemType;
	Handle		itemVal;
	Str255		itemText;
	Rect		itemRect;
	bool		has_cfg;
	long		seekstart;
	long		seekend;

	/* Initialize everything.
	*/
	InitGraf(&qd.thePort);
	InitFonts();
	FlushEvents(everyEvent, 0);
	InitWindows();
	InitMenus();
	TEInit();
	InitDialogs(0L);
	InitCursor();
	MaxApplZone();
	
	/* Get filespecs for all the needed files.
	*/
	filemsg("\pWhere is the TADS Run-time?");
	if (os_askfile("Self-Loader", &origname, 'APPL', false))
		exit(0);
	
	filemsg("\pSelect game file:");
	if (os_askfile("Game File", &gamename, 'TADG', 0))
		exit(0);
	
	filemsg("\pSelect config file (Cancel if none):");
	if (os_askfile("Config File", &cfgname, 'TEXT', 0))
		has_cfg = false;
	else
		has_cfg = true;
	
	nofilemsg();
	if (os_askfile("Write program file to:", &progname, 'APPL', true))
		exit(0);
	
	/* Copy run-time to new application.
	*/
	err = FSpCreate(&progname, 'TADR', 'APPL', smSystemScript);
	if (err != noErr && err != dupFNErr)
		errexit("creating program file", err);
	
	// -- copy data fork.
	oref.open_df(&origname, fsRdPerm, "opening TADS run-time data fork");
	pref.open_df(&progname, fsWrPerm, "opening target data fork");
	SetEOF(pref.refnum(), 0L);
	copy_fork(oref, pref, "run-time data fork");
	GetFPos(pref.refnum(), &seekend);  // get size of data fork; we'll need it in a bit.
	oref.close();
	pref.close();
	
	// -- copy resource fork.
	oref.open_rf(&origname, fsRdPerm, "opening TADS run-time resource fork");
	pref.open_rf(&progname, fsWrPerm, "opening target resource fork");
	SetEOF(pref.refnum(), 0L);
	copy_fork(oref, pref, "run-time resource fork");
	oref.close();
	pref.close();

	// -- munge cfrg resource.
	{
		/*
			Because the application might be fat, we have to change any
			uses of the "whole fork" constant in its cfrg resource to the
			actual size of the fork.
		*/
		refnum_t		pres;
		CFragResource **cfrg;
		
		pres = FSpOpenResFile(&progname, fsRdWrPerm);
		if (pres == -1)
			errexit("opening target resource fork", ResError());
		
		cfrg = (CFragResource **) Get1Resource('cfrg', 0);
		if (cfrg) {
			CFragResourceMember	   *member;
			UInt16					i;
			
			HLock((Handle) cfrg);
			for (i = (**cfrg).memberCount, member = &(**cfrg).firstMember;
				 i > 0;
				 i--, member = NextCFragResourceMemberPtr(member))
			{
				if (member->length == kCFragGoesToEOF)
					member->length = seekend - member->offset;
			}
			HUnlock((Handle) cfrg);
			ChangedResource((Handle) cfrg);
		}
		CloseResFile(pres);
	}
	
	/* Copy game file.
	*/
	gref.open_df(&gamename, fsRdPerm, "opening game file");
	if (has_cfg)
		cref.open_df(&cfgname, fsRdPerm, "opening config file");
	
	pref.open_df(&progname, fsRdWrPerm, "opening target application");
	SetFPos(pref.refnum(), fsFromLEOF, 0);		// appending to data fork.
	
	pref.write('MK22', "writing game file");	// "file" header.
	
	pref.write('TGAM', "writing game file");	// game prefix.
	GetFPos(pref.refnum(), &seekstart);			// remember where we are...
	pref.write(0L,     "writing game file");	// ...and write space for offset field; we'll fill it in later.	
	copy_fork(gref, pref, "game file");			// copy the game file itself.
	
	/* go back and write the seek position of the next segment */
	GetFPos(pref.refnum(), &seekend);
	SetFPos(pref.refnum(), fsFromStart, seekstart);
	pref.write(seekend, "writing game file");
	SetFPos(pref.refnum(), fsFromStart, seekend);
	
	/* write the config file block, if desired */
	if (has_cfg)
	{
		/* write the config prefix */
		pref.write('RCFG', "writing config file");
		GetFPos(pref.refnum(), &seekstart);
		pref.write(0L,     "writing config file");
		
		/* write the size of the config data */
		SetFPos(cref.refnum(), fsFromLEOF, 0L);
		GetFPos(cref.refnum(), &seekend);
		SetFPos(cref.refnum(), fsFromStart, 0L);
		pref.write(seekend, "writing config file");

		/* copy the config file */
		copy_fork(cref, pref, "config file");
		
		/* go back and write the seek position of the next segment */
		GetFPos(pref.refnum(), &seekend);
		SetFPos(pref.refnum(), fsFromStart, seekstart);
		pref.write(seekend, "writing config file");
		SetFPos(pref.refnum(), fsFromStart, seekend);
	}
	
	/* done with the data fork */	
	pref.close();
	
		
	FSpGetFInfo(&progname, &myinfo);
	myinfo.fdFlags |= fHasBundle;  /* file has a bundle */
	
	mydlg = GetNewDialog(25832, NULL, (WindowPtr) -1L);
	SelectDialogItemText(mydlg, 3, 0, 32767);
	do
	{
		ModalDialog(NULL, &itemHit);
	} while (itemHit != 1);
	GetDialogItem(mydlg, 3, &itemType, &itemVal, &itemRect);
	GetDialogItemText(itemVal, itemText);
	if (itemText[0] > 4) itemText[0] = 4;

	myinfo.fdCreator = 0;
	memcpy(&myinfo.fdCreator, &itemText[1], itemText[0]);

	FSpSetFInfo(&progname, &myinfo);

	errexit("Success!", 0);
}

