/*
	TADS:
	
	Defines used by portable TADS code.
*/

#define MAC_OS 1

#define TADS_OEM_NAME			"Chris Nebel <c.nebel@apple.com>"
#define TADS_OEM_VERSION		"1"
#define OS_SYSTEM_PATCHSUBLVL	"0"


/*
	DEBUGGING CONTROL:
	
	Change the debugging level by setting BUILD_KIND to one of the three
	defined constants below.
	
	There are two debugging switches used in TADS. Both work on a defined/
	undefined basis -- undefined means false, defined means true.  (The Mac
	OS build will always define them to 1, so #if and #ifdef will work
	equally well.)

	DEBUG:		The main debugging switch for TADS.  You get verbose status
				reporting if this is on.
		
	ASSERTS:	Controls presence of asserts; that is, assert, dprintf, and
				the macros DPRINTF and VERIFY.  Undefined if asserts should
				be used, defined if not.  (Note that only assert may be used
				in portable code.)
	
	(This header sets a third switch, NDEBUG.  It is only used by assert.h
	and mdebug; it should not appear anywhere in the TADS source code.)
*/

#define kDEBUG       0
#define kPRERELEASE  1
#define kRELEASE     2

#define BUILD_KIND  kRELEASE

#if BUILD_KIND == kDEBUG
	// Turn on all debugging.
	#define DEBUG 1
	#define ASSERTS 1
	#pragma global_optimizer off
#elif BUILD_KIND == kPRERELEASE
	// Asserts only.
	#undef  DEBUG
	#define ASSERTS 1
	#pragma global_optimizer on
	#pragma optimization_level 2
#elif BUILD_KIND == kRELEASE
	// No debugging code.
	#undef  DEBUG
	#undef  ASSERTS
	#pragma global_optimizer on
	#pragma optimization_level 2
#elif
	#error "Unknown BUILD_KIND."
#endif

#ifdef ASSERTS
	#undef NDEBUG
#else
	#define NDEBUG 1
#endif


/*
	OTHER:
	
	Defines used in Mac-specific code and for personal use.
*/

#define MANGLE_ISSTANDALONE 0
#define DEBUG_IO 0
