#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/TADS2/mac/DBGUMAC.C,v 1.2 1999/05/17 02:52:18 MJRoberts Exp $";
#endif

/* Copyright (c) 1992 Michael J. Roberts.  All Rights Reserved. */
/*
Name
  dbgumac.c  - debugger user interface for macintosh
Function
  Implements user interface for TADS 2 debugger for Mac
Notes
  None
Modified
  08/22/92 MJRoberts   - creation
*/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <ctype.h>
#include <stdarg.h>

#include <QuickDraw.h>
#include <Menus.h>
#include <Types.h>
#include <Files.h>
#include <Windows.h>
#include <Fonts.h>
#include <TextEdit.h>
#include <Dialogs.h>
#include <Events.h>
#include <Controls.h>
#include <ToolUtils.h>
#include <Lists.h>

#include "os.h"
#include "dbg.h"
#include "tio.h"
#include "obj.h"
#include "prp.h"
#include "tok.h"
#include "run.h"
#include "lin.h"
#include "linf.h"
#include "ossmac.h"
#include "dbgumac.h"

int tdb2 = 1;           /* indicate to os subsystem that this is TDB 2.0 */
#if 0
// -- defined in dummy.c.
int tadsdebug = 0;      /* but not TDB 1.0! */
#endif

extern WindowPtr                myWindow;
extern WindowPtr                dbgWindow;
extern WindowPtr                gameWindow;
extern ControlHandle    svScroll, shScroll, wvScroll, whScroll;
extern MenuHandle               myMenus[];

static char dbg_curstat[81];

#define DBGUHLIN '-'

/* location in source file of top current source window and current line */
static long dbgsrctop;                          /* current display seek location */
static long dbgsrcloc;                          /* current source line in file */
static long dbgsrchome;                         /* seek location of "home" screen */

/* lindef for current source file */
static osfildef *dbgsrcfp;                      /* current file displayed in source window */
static lindef *dbgsrclin;                       /* lindef for currently displayed file */
static lindef *dbgsrclinhome;           /* lindef of "home" file */

/* private breakpoint data - maintained with dbgbpeach() */
static int dbgbpcnt;                                    /* number of breakpoints set */
typedef struct dbgubpdef dbgubpdef;
static struct dbgubpdef
{
        int             dbgubpid;                               /* line source id */
        long    dbgubpseek;                             /* seek position of breakpoint */
}
dbgbps[DBGBPMAX];                                       /* seek locations of breakpoints */

extern WindowPtr        srcWindow;
extern WindowPtr        watchWindow;
WindowRecord            wwRecord;

static dbgcxdef *dbgctx;

int watch_selection = -1;
static int watch_count;

static void dbgptf(char *fmt, ...)
{
        char    buf[128];
        va_list ap;
        
        va_start(ap, fmt);
        vsprintf(buf, fmt, ap);
        va_end(ap);
        os_putln(buf);
}

/* forward declarations */
static void dbgugetbp(dbgcxdef *ctx);


/*
 *   Debugger user interface initialization, phase two.  TADS calls this
 *   routine during startup, after read the .GAM file.  The debugger user
 *   interface code can perform any required initialization that depends
 *   on the .GAM file having been read.  
 */
void dbguini2(dbgcxdef *ctx)
{
        #pragma unused(ctx)
}

/*
 *   Find a source file.  origname is the name of the source file as it
 *   appears in the game's debugging information; this routine should
 *   figure out where the file actually is, and put the fully-qualified
 *   path to the file in fullname.  The run-time calls this after it
 *   exhausts all of its other methods of finding a source file (such as
 *   searching the include path).
 *   
 *   Return true if the source file should be considered valid, false if
 *   not.  Most implementations will simply return true if the file was
 *   found, false if not; however, this approach will cause the debugger
 *   to terminate with an error at start-up if the user hasn't set up the
 *   debugger's include path correctly before running the debugger.  Some
 *   implementations, in particular GUI implementations, may wish to wait
 *   to find a file until the file is actually needed, rather than pester
 *   the user with file search dialogs repeatedly at start-up.
 *   
 *   must_find_file specifies how to respond if we can't find the file.
 *   If must_find_file is true, we should always return false if we can't
 *   find the file.  If must_find_file is false, however, we can
 *   optionally return true even if we can't find the file.  Doing so
 *   indicates that the debugger UI will defer locating the file until it
 *   is actually needed.
 *   
 *   If this routine returns true without actually finding the file, it
 *   should set fullname[0] to '\0' to indicate that fullname doesn't
 *   contain a valid filename.  
 */
int dbgu_find_src(char *origname, int origlen,
                  char *fullname, size_t full_len, int must_find_file)
{
        #pragma unused(origname, origlen, fullname, full_len, must_find_file)
        
        return false;
}

/*
 *   Determine if the debugger can resume from a run-time error.  This
 *   reflects the capabilities of the user interface of the debugger.  In
 *   particular, if the UI provides a way to change the instruction
 *   pointer, then the debugger can resume from an error, since the user
 *   can always move past the run-time error and continue execution.  If
 *   the UI doesn't let the user change the instruction pointer, resuming
 *   from an error won't work, since the program will keep hitting the
 *   same error and re-entering the debugger.  If this returns false, the
 *   run-time will trap to the debugger on an error, but will simply abort
 *   the current command when the debugger returns.  If this returns true,
 *   the run-time will trap to the debugger on an error with the
 *   instruction pointer set back to the start of the line containing the
 *   error, and will thus re-try the same line of code when the debugger
 *   returns, unless the debugger explicitly moves the instruction pointer
 *   before returning.  
 */
int dbgu_err_resume(dbgcxdef *ctx)
{
        #pragma unused(ctx)
        
        return false;
}


/*
 *   Debugger UI - quitting game.  The runtime calls this routine just
 *   before the play loop is about to terminate after the game code has
 *   called the "quit" built-in function.  If the debugger wants, it can
 *   take control here (just as with dbgucmd()) for as long as it wants.
 *   If the debugger wants to restart the game, it should call bifrst().
 *   If this routine returns without signalling a RUN_RESTART error, TADS
 *   will terminate.  If a RUN_RESTART error is signalled, TADS will
 *   resume the play loop.  
 */
void dbguquitting(dbgcxdef *ctx)
{
        #pragma unused(ctx)
}

/* 
 *   debugger user interface termination - this routine is called when the
 *   debugger is about to terminate, so that the user interface can close
 *   itself down (close windows, release memory, etc) 
 */
void dbguterm(dbgcxdef *ctx)
{
        #pragma unused(ctx)
}


/* move back or forward by one or more lines within a file */
static char dbgbuf[1024];

static void dbgmove(osfildef *fp, int cur, int *destlinep)
{
        int             dest = *destlinep;
        long    pos = osfpos(fp);
        long    prvpos;
        char   *p;
        size_t  max;
        int             backup;
        
        if (cur < dest)
        {
                while (cur != dest)
                {
                        max = fread(dbgbuf, 1, sizeof(dbgbuf), fp);
                        if (max == 0) break;
                        for (p = dbgbuf ; p < dbgbuf + max && cur != dest ; ++p, ++pos)
                                if (*p == '\n' || *p == '\r') ++cur;
                }
        }
        else
        {
                ++cur;
                while (cur != dest)
                {
                        backup = (pos > sizeof(dbgbuf) ? sizeof(dbgbuf) : pos);
                        prvpos = pos - backup;
                        osfseek(fp, prvpos, OSFSK_SET);
                        max = fread(dbgbuf, 1, backup, fp);
                        if (max == 0) break;
                        p = dbgbuf + max;
                        do
                        {
                                --p;
                                --pos;
                                if (*p == '\n' || *p == '\r') --cur;
                        } while (p > dbgbuf && cur != dest);
                        if (prvpos == 0) break;
                }
                if (pos > 0) ++pos;
                else cur = 1;
        }
        
        osfseek(fp, pos, OSFSK_SET);
        *destlinep = cur;
}

/* show current source page */
static void dbgsrcshow(lindef *lin, osfildef *fp, long top, long loc)
{
        int                     y;
        char            outbuf[128];
        long            curloc;
        int                     i;
        dbgubpdef  *bp;
        osfildef   *curfp = dbgsrcfp;
        int                     thisid = lin->linid;
        int                     ht;
        FontInfo        f;
        int                     delta;
        long            initloc = osfpos(fp);
        int                     hofs = GetControlValue(shScroll);

        TextFont(4);
        TextSize(9);
        GetFontInfo(&f);
        
        if (!fp)
        {
                MoveTo(9, 9);
                DrawText((Ptr)"Source file not available", 0, 25);
                return;
        }
        
        ht = srcWindow->portRect.bottom - srcWindow->portRect.top - kScrollBarWidth - 1;
        delta = f.ascent + f.descent + f.leading;
        
        for (y = delta + 3 ; y < ht - 3 ; y += delta)
        {
                curloc = osfpos(fp);
                
                /* special display if this is the currently executing line */
                if (curloc == loc && fp == curfp)
                {
                        MoveTo(7, y);
                        TextFont(23);
                        DrawChar(0xde);
                }
                
                /* special display if there is a breakpoint at this line */
                for (i = dbgbpcnt, bp = dbgbps ; i ; ++bp, --i)
                {
                        if (bp->dbgubpseek == curloc && bp->dbgubpid == thisid)
                        {
                                MoveTo(0, y);
                                TextFont(0);
                                DrawChar(0x13);
                                break;
                        }
                }

                /* display the line itself */
                if (!osfgets(outbuf, sizeof(outbuf), fp))
                        break;
                else
                {
                        int             l;
                        char   *p = outbuf;
                        
                        MoveTo(15, y);
                        l = strlen(outbuf);
                        if (l && (outbuf[l-1] == '\n' || outbuf[l-1] == '\r')) --l;
                        if (l > hofs)
                        {
                                p += hofs;
                                TextFont(4);
                                TextSize(9);
                                DrawText((Ptr)p, 0, l - hofs);
                        }
                }
        }
        
        /* set the vertical scroll thumb */
        osfseek(fp, 0L, OSFSK_END);
        SetControlMinimum(svScroll, 0);
        SetControlMaximum(svScroll, (int)(osfpos(fp) / 256));
        SetControlValue(svScroll, (int)(initloc / 256));
}

/* callback for os layer to update source screen */
void dbgsrcupd(void)
{
        Rect r;
        
        r = srcWindow->portRect;
        r.right -= kScrollBarWidth + 1;
        r.bottom -= kScrollBarWidth + 1;
        ClipRect(&r);
        
        osfseek(dbgsrcfp, dbgsrctop, OSFSK_SET);
        dbgsrcshow(dbgsrclin, dbgsrcfp, dbgsrctop, dbgsrcloc);
        
        r.left = r.top = 0;
        r.right = r.bottom = 32767;
        ClipRect(&r);
}

/* invert color of a line in watch window */
static void dbgwxsel(int lin, int lht)
{
        Rect r;
        int  newbot;
        
        lin -= GetControlValue(wvScroll);
        if (lin < 0) return;
        r = watchWindow->portRect;
        r.top = (lin * lht) + 5;
        newbot = r.top + lht;
        if (newbot > r.bottom - 15) newbot = r.bottom - 15;
        r.bottom = newbot;
        r.left += 2;
        r.right -= 15;
        InvertRect(&r);
}

/* context for callback to draw text into watch window */
struct wxcxdef
{
        int wxcxdelta;
        int wxcxy;
        int wxcxcurx;
        int wxcxleftx;
        int wxcxcury;
        int wxcxtopy;
};
typedef struct wxcxdef wxcxdef;

/* callback to draw text into watch window */
static void dbgwdisp(void /*wxcxdef*/ *ctx_v, char *str, int strl)
{
        wxcxdef *ctx = (wxcxdef *) ctx_v;
        
        if (strl == 1 && (*str == '\n' || *str == '\r'))
        {
                /* move to next line in window if anything's shown yet */
                if (ctx->wxcxcury >= ctx->wxcxtopy)
                {
                        ctx->wxcxy += ctx->wxcxdelta;
                        MoveTo(3, ctx->wxcxy);
                }
                
                /* move to left of new line in absolute destination view */
                ctx->wxcxcury++;
                ctx->wxcxcurx = 0;
        }
        else
        {
                /* ignore if not at first line showing in window yet */
                if (ctx->wxcxcury < ctx->wxcxtopy) return;
                
                /* ignore if not at leftmost char in window yet */
                if (strl + ctx->wxcxcurx < ctx->wxcxleftx)
                        ctx->wxcxcurx += strl;
                else
                {
                        if (ctx->wxcxcurx < ctx->wxcxleftx)
                        {
                                str += ctx->wxcxleftx - ctx->wxcxcurx;
                                strl -= ctx->wxcxleftx - ctx->wxcxcurx;
                                ctx->wxcxcurx = ctx->wxcxleftx;
                        }
                        DrawText((Ptr)str, 0, strl);
                }
        }
}

/* update watch window with current clipping rectangle */
static void dbgwatchupd0(void)
{
        FontInfo        f;
        wxcxdef         ctx;
        
        GetFontInfo(&f);
        ctx.wxcxdelta = f.ascent + f.descent + f.leading;
        ctx.wxcxy = ctx.wxcxdelta + 3;
        
        ctx.wxcxcury = 0;
        ctx.wxcxtopy = GetControlValue(wvScroll);
        ctx.wxcxcurx = 0;
        ctx.wxcxleftx = GetControlValue(whScroll);
        
        MoveTo(3, ctx.wxcxy);
        dbgwxupd(dbgctx, dbgwdisp, &ctx);
        
        /* hilite selection line */
        if (watch_selection != -1)
                dbgwxsel(watch_selection, ctx.wxcxdelta);
}

/* update watch window */
void dbgwatchupd(void)
{
        Rect    r;
        
        r = watchWindow->portRect;
        r.right -= 15;
        r.bottom -= 15;
        ClipRect(&r);

        dbgwatchupd0();

        r.left = r.top = 0;
        r.bottom = r.right = 32767;
        ClipRect(&r);
}

/* invalidate source window */
static void dbginval0(int all /* true ==> inval entire window; false ==> only bp area */)
{
        GrafPtr savePort;
        Rect    r;

        GetPort(&savePort);
        SetPort((GrafPtr)srcWindow);
        r = srcWindow->portRect;
        r.bottom -= kScrollBarWidth + 1;
        if (all)
                r.right -= kScrollBarWidth + 1;
        else
                r.right = r.left + 7;
        InvalRect(&r);
        SetPort(savePort);
}

/* invalidate entire source window */
void dbginval(void)
{
        dbginval0(1);
}

/* invalidate watch window */
void dbgwxinval(void)
{
        GrafPtr savePort;
        Rect    r;

        if (!watchWindow) return;
        GetPort(&savePort);
        SetPort((GrafPtr)watchWindow);
        r = watchWindow->portRect;
        r.bottom -= kScrollBarWidth + 1;
        r.right -= kScrollBarWidth + 1;
        InvalRect(&r);
        SetPort(savePort);
}

/* context structure for watch deletion identification */
struct dbgwdctx
{
        int dbgwdcnt;   /* count of items seen so far */
        int dbgwdrdy;   /* newline just encountered */
        int dbgwdid;    /* id of selected watchpoint */
};
typedef struct dbgwdctx dbgwdctx;

/* callback for watch deletion identification */
static void dbgwxdcb(void /*dbgwdctx*/ *ctx_v, char *str, int strl)
{
        dbgwdctx *ctx = (dbgwdctx *) ctx_v;
        
        if (*str == '[' && ctx->dbgwdrdy)
        {
                if (ctx->dbgwdcnt == watch_selection)
                        ctx->dbgwdid = atoi(str + 1);
                ctx->dbgwdcnt++;
        }
        ctx->dbgwdrdy = (*str == '\n' || *str == '\r');
}

/* adjust size of watch window */
void dbgwxgrow(void)
{
        Rect            r;
        GrafPtr         savePort;
        FontInfo        f;
        int                     ht;
        
        GetPort(&savePort);
        SetPort(watchWindow);
        
        r = watchWindow->portRect;
        GetFontInfo(&f);
        ht = (r.bottom - r.top - 20) / (f.ascent + f.descent + f.leading);
        if (watch_count > ht)
        {
                SetControlMaximum(wvScroll, watch_count - ht);
        }
        else
        {
                SetControlMaximum(wvScroll, 0);
                SetControlValue(wvScroll, 0);
        }
        
        SetPort(savePort);
}

/* add or delete watch expressions */
static void dbguwxadd(dbgcxdef *ctx, int cnt)
{
        if (watch_count == 0 && cnt > 0)
        {
                GrafPtr savePort;
                Rect    vScrollRect;
                
                GetPort(&savePort);
                watchWindow = GetNewWindow(25553, &wwRecord, (WindowPtr) -1L);
                SetPort(watchWindow);
                
                vScrollRect = watchWindow->portRect;
                vScrollRect.top -= 1;
                vScrollRect.left = vScrollRect.right - 15;
                vScrollRect.right += 1;
                vScrollRect.bottom -= 14;
                wvScroll = NewControl(watchWindow, &vScrollRect, "\p", 0, 0, 0, 0, scrollBarProc, 0L);
                HideControl(svScroll);
                
                vScrollRect = watchWindow->portRect;
                vScrollRect.top = vScrollRect.bottom - 15;
                vScrollRect.bottom += 1;
                vScrollRect.left -= 1;
                vScrollRect.right -= 14;
                whScroll = NewControl(watchWindow, &vScrollRect, "\p", 0, 0, 0, 1024, scrollBarProc, 0L);
                HideControl(shScroll);
                
                TextFont(4);
                TextSize(9);
                SetPort(savePort);
        }
        
        watch_count += cnt;
        
        if (watch_count == 0)
        {
                DisposeControl(wvScroll);
                DisposeControl(whScroll);
                DisposeWindow(watchWindow);
                watchWindow = 0;
                wvScroll = whScroll = 0;
        }
        else
        {
                dbgwxgrow();
                dbgwxinval();
        }
}

void dbgwxclr(void)
{
        dbgwdctx ctx;
        
        if (watch_selection == -1) return;
        
        /* call callback to learn identity of selected watch */
        ctx.dbgwdcnt = 0;
        ctx.dbgwdid = -1;
        ctx.dbgwdrdy = 1;
        dbgwxupd(dbgctx, dbgwxdcb, &ctx);
        
        /* delete selected watch, if we found a match */
        if (ctx.dbgwdid != -1)
        {
                dbgwxdel(dbgctx, ctx.dbgwdid);
                dbguwxadd(dbgctx, -1);
                watch_selection = -1;
                dbgwxinval();
        }
}

/* adjust source window for new thumb position */
void srcAdjust(void)
{
        long newpos = (long)GetControlValue(svScroll) * 256L;
        char buf[128];
        
        /* seek to approximate position, then find end of line there */
        osfseek(dbgsrcfp, newpos, OSFSK_SET);
        if (newpos) osfgets(buf, (int)sizeof(buf), dbgsrcfp);
        dbgsrctop = osfpos(dbgsrcfp);
        dbginval();
}

/* show the current line */
static void dbgulsho(dbgcxdef *ctx, uchar *buf /* line information retrieved by dbglget */)
{
        lindef     *lin;
        int                     cur;
        int                     dest;
        osfildef   *fp;
        long            loc;
        int                     ht;
        FontInfo        f;
        
        if (!dbgsrclin && dbgsrcfp) fclose(dbgsrcfp);
        
        GetFontInfo(&f);
        ht = (srcWindow->portRect.bottom - srcWindow->portRect.top)
                 / (f.ascent + f.descent + f.leading);
        
        /* search for a line source with this id */
        for (lin = ctx->dbgcxlin ; lin ; lin = lin->linnxt)
                if (lin->linid == *buf) break;
        dbgswitch(&dbgsrclin, lin);
        dbgsrclinhome = lin;
        
        if (!lin)
        {
                return;
        }

        /* seek to line in file, then back up by half a screen to center line */
        fp = dbgsrcfp = ((linfdef *)lin)->linffp;
        loc = osrp4(buf + 1);
        osfseek(fp, loc, OSFSK_SET);
        cur = ht / 2;
        dest = 1;
        dbgmove(fp, cur, &dest);
        dbgsrctop = dbgsrchome = osfpos(fp);
        dbgsrcloc = loc;
        dbginval();
        
        #ifdef NEVER
        {
                char   *fnam;
                int             i;
        
                /* update status line to show current file */
                for (i = 0 ; i < 5 ; ++i) dbg_curstat[i] = DBGUHLIN;
                
                fnam = ((linfdef *)lin)->linfnam;
                sprintf(&dbg_curstat[5], "%s    ", fnam);
                dbgwhere(ctx, &dbg_curstat[strlen(dbg_curstat)]);
        
                for (i = strlen(dbg_curstat); i <= dbgstawin.oswx2 - dbgstawin.oswx1; ++i)
                        dbg_curstat[i] = DBGUHLIN;
                dbg_curstat[i] = '\0';
        
                /* actually display the status line now */
                dbgustat(ctx);
        }
        #endif /* NEVER */
}

/* scroll the watch window */
void dbgwxScroll(ControlHandle theControl, int theCode)
{
        int                     scrollAmt;
        int                     ht;
        FontInfo        f;
        Rect            r;
        GrafPtr         savePort;
        RgnHandle       rgn;
        int                     lht;

        GetPort(&savePort);
        SetPort(watchWindow);

        if (theControl == wvScroll)
        {
                GetFontInfo(&f);
                lht = (f.ascent + f.descent + f.leading);
                ht = (watchWindow->portRect.bottom - watchWindow->portRect.top) / lht;
        }
        else
        {
                ht = (watchWindow->portRect.right - watchWindow->portRect.left - 15) / 6;
        }
        
        if (theCode == 0) return;
        
        switch (theCode)
        {
        case kControlUpButtonPart: 
                scrollAmt = -1;
                break;
        case kControlDownButtonPart: 
                scrollAmt = 1;
                break;
        case kControlPageUpPart: 
                scrollAmt = -ht;
                break;
        case kControlPageDownPart: 
                scrollAmt = ht;
                break;
        }
        
        r = watchWindow->portRect;
        r.right -= kScrollBarWidth;
        r.bottom -= kScrollBarWidth;
        
        if ((GetControlValue(theControl) == GetControlMinimum(theControl) && scrollAmt < 0)
                || (GetControlValue(theControl) == GetControlMaximum(theControl) && scrollAmt > 0))
                return;
        
        SetControlValue(theControl, GetControlValue(theControl) + scrollAmt);

        if (theControl == wvScroll)
        {
                if (scrollAmt == 1 || scrollAmt == -1)
                {
                        r.top += 5;
                        rgn = NewRgn();
                        ScrollRect(&r, 0, -(scrollAmt * lht), rgn);
                        SetClip(rgn);
                        DisposeRgn(rgn);
                }
        }
        else
        {
                if (scrollAmt == 1 || scrollAmt == -1)
                {
                        r.top -= 1;
                        r.bottom += 1;
                        r.left += 3;
                        rgn = NewRgn();
                        ScrollRect(&r, -6 * scrollAmt, 0, rgn);
                        SetClip(rgn);
                        DisposeRgn(rgn);
                }
        }
        
        if (scrollAmt != 1 && scrollAmt != -1)
        {
                r.right += 1;
                EraseRect(&r);
                r.right -= 1;
                ClipRect(&r);
        }
                
        dbgwatchupd0();
        
        r.left = r.top = 0;
        r.right = r.bottom = 32767;
        ClipRect(&r);
        DrawControls(watchWindow);
        SetPort(savePort);
}

/* scroll source window */
void srcScroll(ControlHandle theControl, int theCode)
{
        int                     scrollAmt;
        int                     ht;
        FontInfo        f;
        int                     cur;
        int                     dest;
        Rect            r;
        GrafPtr         savePort;
        RgnHandle       rgn;
        int                     lht;

        GetPort(&savePort);
        SetPort(srcWindow);

        if (theControl == svScroll)
        {
                GetFontInfo(&f);
                lht = (f.ascent + f.descent + f.leading);
                ht = (srcWindow->portRect.bottom - srcWindow->portRect.top) / lht;
        }
        else
        {
                ht = (srcWindow->portRect.right - srcWindow->portRect.left - 15) / 6;
        }
        
        if (theCode == 0) return;
        
        switch (theCode)
        {
        case kControlUpButtonPart: 
                scrollAmt = -1;
                break;
        case kControlDownButtonPart: 
                scrollAmt = 1;
                break;
        case kControlPageUpPart: 
                scrollAmt = -ht;
                break;
        case kControlPageDownPart: 
                scrollAmt = ht;
                break;
        }
        
        osfseek(dbgsrcfp, dbgsrctop, OSFSK_SET);
        r = srcWindow->portRect;
        r.right -= kScrollBarWidth + 1;
        r.bottom -= kScrollBarWidth + 1;
        
        if (theControl == svScroll && scrollAmt < 0)
        {
                if (osfpos(dbgsrcfp) == 0) return;
        }
        else if ((GetControlValue(theControl) == GetControlMinimum(theControl) && scrollAmt < 0)
                         || (GetControlValue(theControl) == GetControlMaximum(theControl) && scrollAmt > 0))
                return;
        
        if (theControl == shScroll)
                SetControlValue(theControl, GetControlValue(theControl) + scrollAmt);

        if (theControl == svScroll)
        {
                cur = dest = 1000;
                dest += scrollAmt;
                dbgmove(dbgsrcfp, cur, &dest);
                dbgsrctop = osfpos(dbgsrcfp);
        
                if (scrollAmt == 1 || scrollAmt == -1)
                {
                        r.bottom = (r.bottom - r.top + 4) / lht;
                        r.bottom *= lht;
                        r.bottom += r.top - 5;
                        r.top += 5;
                        rgn = NewRgn();
                        ScrollRect(&r, 0, -(scrollAmt * lht), rgn);
                        SetClip(rgn);
                        DisposeRgn(rgn);
                }
        }
        else
        {
                if (scrollAmt == 1 || scrollAmt == -1)
                {
                        r.top -= 1;
                        r.bottom += 1;
                        r.left += 15;
                        rgn = NewRgn();
                        ScrollRect(&r, -6 * scrollAmt, 0, rgn);
                        SetClip(rgn);
                        DisposeRgn(rgn);
                }
        }
        
        if (scrollAmt != 1 && scrollAmt != -1)
        {
                r.right += 1;
                EraseRect(&r);
                r.right -= 1;
                ClipRect(&r);
        }

        dbgsrcshow(dbgsrclin, dbgsrcfp, dbgsrctop, dbgsrcloc);
        
        r.left = r.top = 0;
        r.right = r.bottom = 32767;
        ClipRect(&r);
        DrawControls(srcWindow);
        SetPort(savePort);
}

/* show default button */
static void dfltbtn(DialogPtr dlg)
{
        Rect    r;
        
        SetPort((GrafPtr)dlg);
        GetControlItemRect(dlg, 1, &r);
        PenSize(3, 3);
        InsetRect(&r, -4, -4);
        FrameRoundRect(&r, 16, 16);
}

/* handle a mouse click inside the source window */
void dbgContent(Point where, EventRecord *evt)
{
        GrafPtr         savePort;
        FontInfo        f;
        int                     lht;
        int                     linofs;
        char       *cond = NULL;
        char            condbuf[128];
        long            linpos;
        extern int      os0td_quitting;
        char            posbuf[4];
        
        if (os0td_quitting) return;
        
        GetPort(&savePort);
        SetPort(srcWindow);
        
        GetFontInfo(&f);
        lht = (f.ascent + f.descent + f.leading);
        
        where.v += 4;
        linofs = where.v / lht;
        
        if (dbgsrclin)
        {
                objnum  objn;
                uint    ofs;
                int             newbp;
                char    nambuf[128];
                int             len;
                
                /* find seek location of selected line */
                osfseek(dbgsrcfp, dbgsrctop, OSFSK_SET);
                while (linofs--)
                {
                        linpos = osfpos(dbgsrcfp);
                        if (!osfgets(dbgbuf, sizeof(dbgbuf), dbgsrcfp))
                                goto quit;
                }

                if (evt->modifiers & optionKey)
                {
                        DialogPtr       mydlg;
                        short           itemHit;
                        Str255          textval;
                        
                        mydlg = GetNewDialog(25121, NULL, (WindowPtr) -1L);
                        dfltbtn(mydlg);

                        do
                        {
                                ModalDialog(NULL, &itemHit);
                        } while (itemHit != 1 && itemHit != 2);
                        if (itemHit == 1)
                        {
                                GetControlItemTitle(mydlg, 4, textval);
                                StringCopy_cp(condbuf, textval, 128);
                                cond = condbuf;
                        }
                        DisposeDialog(mydlg);
                }
                
                oswp4(posbuf, linpos);
                linffind(dbgsrclin, posbuf, &objn, &ofs);
                
                memcpy(nambuf, "<<>> ", (size_t)4);
                len = dbgnam(dbgctx, nambuf+4, TOKSTOBJ, objn);
                if (!memcmp(nambuf+4, "<UNKNOWN>", (size_t)len))
                        len = dbgnam(dbgctx, nambuf+4, TOKSTFUNC, objn);
                if (cond)
                {
                        memcpy(nambuf + 4 + len, " when ", (size_t)6);
                        strcpy(nambuf + 10 + len, cond);
                }
                else
                        nambuf[len + 4] = '\0';
                                        
                (void) dbgbpat(dbgctx, objn, MCMONINV, ofs, &newbp, nambuf, TRUE, cond, NULL);
                dbgugetbp(dbgctx);
                dbginval0(0);
        }

quit:
        SetPort(savePort);
}

/* handle a click in watch window */
void dbgwxcont(Point where, EventRecord *evt)
{
        GrafPtr         savePort;
        FontInfo        f;
        int                     lht;
        int                     linofs;
        Rect            r;
        
        GetPort(&savePort);
        SetPort(watchWindow);
        
        GetFontInfo(&f);
        lht = (f.ascent + f.descent + f.leading);

        /* ignore clicks in the scroll bars */
        r = watchWindow->portRect;
        if (where.v > r.bottom - r.top - 15
                || where.h > r.right - r.left - 15)
                return;
        
        if (where.v > 3) where.v -= 4;
        linofs = where.v / lht;
        linofs += GetControlValue(wvScroll);
        
        /* turn off old watch selection, if any */
        if (watch_selection != -1)
        {
                dbgwxsel(watch_selection, lht);
                watch_selection = -1;
        }
        
        /* highlight new selection, if it's within the list */
        if (linofs < watch_count)
        {
                watch_selection = linofs;
                dbgwxsel(watch_selection, lht);
        }
        
quit:
        SetPort(savePort);
}

/* callback for dbgugetp */
static void dbgubp(void /*dbgcxdef*/ *ctx_v, int id, uchar *buf, uint _u)
{
        dbgubpdef *bp = &dbgbps[dbgbpcnt++];

        bp->dbgubpid = id;
        bp->dbgubpseek = osrp4(buf);
}

/* reset private breakpoint data - call after setting/clearing bp */
static void dbgugetbp(dbgcxdef *ctx)
{
        dbgbpcnt = 0;
        dbgbpeach(ctx, dbgubp, ctx);
        if (dbgsrclin)
        {
                osfildef *fp = ((linfdef *)dbgsrclin)->linffp;
                if (fp) dbginval0(0);
        }
}

/* display callback */
static void dbgudisp(dvoid *ctx, char *str, int strl)
{
        WindowPtr savewin;
        
        savewin = myWindow;
        myWindow = dbgWindow;
        dbgptf("%.*s", strl, str);
        myWindow = savewin;
}

/* search for a string in a line source */
static void dbgusearch(lindef *lin, osfildef *fp, long *startpos,
                long *toppos /* location of top of screen */, char *str)
{
        int                     dest;
        int                     ht;
        FontInfo        f;
        
        GetFontInfo(&f);
        ht = (srcWindow->portRect.bottom - srcWindow->portRect.top)
                 / (f.ascent + f.descent + f.leading);
                
        /* search is case-insensitive - convert search string to lowercase */
        strlwr(str);

        /* seek to starting location, and read and discard current line */
        osfseek(fp, *startpos, OSFSK_SET);
        if (!osfgets(dbgbuf, sizeof(dbgbuf), fp)) return;
        
        for (;;)
        {
                *startpos = osfpos(fp); /* record position of current line */
                if (!osfgets(dbgbuf, sizeof(dbgbuf), fp)) break;
                strlwr(dbgbuf);
                if (strstr(dbgbuf, str)) break;
        }
        
        /* move search expression to middle of screen */
        osfseek(fp, *startpos, OSFSK_SET);
        dest = 1;
        dbgmove(fp, ht / 2, &dest);
        *toppos = osfpos(fp);
        dbginval();
}

static void dbgucmd1(dbgcxdef *ctx)
{
        uchar           cmdbuf[128];
        char       *buf;
        int                     err;
        int                     disable;
        int                     bpnum;
        int                     i;
        lindef     *lin;
        static char     lastsearch[128];
        static long     lastpos;
        int                     level = ctx->dbgcxfcn - 1;

        dbglget(ctx, cmdbuf);           /* get information on current line */
        dbgulsho(ctx, cmdbuf);          /* display current line */
        dbgwxinval();                           /* update watch expression values */

        for (;;)
        {
                dbgptf("tdb>");
                os_gets(cmdbuf, 128);
                for (buf = (char *) cmdbuf; *buf && isspace(*buf); ++buf) ;
                                
                if (isupper(buf[0])) buf[0] = tolower(buf[0]);
                switch(buf[0])
                {
                        case 'g':
                                ctx->dbgcxflg &= ~(DBGCXFSS + DBGCXFSO);
                                return;
                                
                        case 't':
                                ctx->dbgcxflg |= DBGCXFSS;
                                ctx->dbgcxflg &= ~DBGCXFSO;
                                return;
                                
                        case 'p':
                                ctx->dbgcxsof = ctx->dbgcxdep; /* step until back at this depth */
                                ctx->dbgcxflg |= (DBGCXFSS + DBGCXFSO);
                                return;
                                
                        case 'c':
                                if (isupper(buf[1])) buf[1] = tolower(buf[1]);
                                switch(buf[1])
                                {
                                        case '+':
                                                ctx->dbgcxflg |= DBGCXFTRC;
                                                dbgptf("Call tracing enabled\n");
                                                SetItemMark(myMenus[3], 5,
                                                                        (dbgctx->dbgcxflg & DBGCXFTRC) ? 0x12 : 0x00);
                                                break;
                
                                        case '-':
                                                ctx->dbgcxflg &= ~DBGCXFTRC;
                                                dbgptf("Call tracing disabled\n");
                                                SetItemMark(myMenus[3], 5,
                                                                        (dbgctx->dbgcxflg & DBGCXFTRC) ? 0x12 : 0x00);
                                                break;
                
                                        case 'c':
                                                ctx->dbgcxhstf = 0;
                                                dbgptf("Call trace log cleared\n");
                                                break;
                
                                        case '\0':
                                        {
                                                char *p;
                                                char *endp;
                
                                                for (p = ctx->dbgcxhstp, endp = p + ctx->dbgcxhstf ;
                                                         p < endp ; p += strlen(p) + 1)
                                                {
                                                        dbgptf("%s\n", p);
                                                }
                                        }
                                        break;
                                }
                                break;
        
                        case 'e':
                                myWindow = gameWindow;  /* send dstring output to game window */
                                err = dbgeval(ctx, buf+1, dbgudisp, (void *)0, level, false);
                                os_flush();
                                myWindow = dbgWindow;   /* back to command window */
                                if (err) errlog(ctx->dbgcxerr, err);
                                dbgwxinval();
                                break;
                                
                        case 'b':
                                if (isupper(buf[1])) buf[1] = tolower(buf[1]);
                                switch(buf[1])
                                {
                                        case 'p':
                                                if ((err = dbgbpset(ctx, buf+2, &bpnum)) != 0)
                                                        errlog(ctx->dbgcxerr, err);
                                                else
                                                {
                                                        dbgptf("breakpoint %d set\n", bpnum);
                                                        dbgugetbp(ctx);
                                                }
                                                break;
                                                
                                        case 'c':
                                                if ((err = dbgbpdel(ctx, atoi(buf + 2))) != 0)
                                                        errlog(ctx->dbgcxerr, err);
                                                else
                                                {
                                                        dbgptf("breakpoint cleared\n");
                                                        dbgugetbp(ctx);
                                                }
                                                break;
                
                                        case 'd':
                                                disable = TRUE;
                                                goto bp_enable_or_disable;
                                        case 'e':
                                                disable = FALSE;
                                        bp_enable_or_disable:
                                                if ((err = dbgbpdis(ctx, atoi(buf+2), disable)) != 0)
                                                        errlog(ctx->dbgcxerr, err);
                                                else
                                                        dbgptf("breakpoint %sabled\n", disable ? "dis" : "en");
                                                break;
                                                
                                        case 'l':
                                                dbgbplist(ctx, dbgudisp, (void *)0);
                                                break;
                
                                        default:
                                                dbgptf("invalid breakpoint command\n");
                                                break;
                                }
                                break;
                                
                        case 'w':
                                if (isupper(buf[1])) buf[1] = tolower(buf[1]);
                                switch(buf[1])
                                {
                                        case 's':
                                                if ((err = dbgwxset(ctx, buf+2, &bpnum, level)) != 0)
                                                        errlog(ctx->dbgcxerr, err);
                                                else
                                                        dbguwxadd(ctx, 1);      /* adjust windows for a new wx */
                                                break;
                                                
                                        case 'd':
                                                if ((err = dbgwxdel(ctx, atoi(buf + 2))) != 0)
                                                        errlog(ctx->dbgcxerr, err);
                                                else
                                                        dbguwxadd(ctx, -1);     /* adjust windows for one less wx */
                                                break;
                                                
                                        default:
                                                dbgptf("invalid watch command\n");
                                                break;
                                }
                                break;
                                
                        case 'f':
                                if (isupper(buf[1])) buf[1] = tolower(buf[1]);
                                switch(buf[1])
                                {
                                        case 'l':
                                                for (i = 1, lin = ctx->dbgcxlin ; lin ;
                                                         lin = lin->linnxt, ++i)
                                                {
                                                        dbgptf("#%-2d   %s", i, ((linfdef *)lin)->linfnam);
                                                        #ifdef NEVER
                                                        if (!((linfdef *)lin)->linffp)
                                                                dbgptf("(NOT FOUND)");
                                                        #endif /* NEVER */
                                                        dbgptf("\n");
                                                }
                                                break;
                                                
                                        case 'v':
                                                if (atoi(buf + 2) == 0)
                                                {
                                                        osfildef *fp;
                                                        
                                                        /* file by name - read the filename */
                                                        for (buf += 2; isspace(*buf); ++buf) ;
                                                        if (!(fp = osfoprt(buf, OSFTTEXT)))
                                                                dbgptf("file not found\n");
                                                        else
                                                        {
                                                                if (!dbgsrclin && dbgsrcfp) fclose(dbgsrcfp);
                                                                dbgswitch(&dbgsrclin, (lindef *)0);
                                                                dbgsrcfp = fp;
                                                                dbgsrctop = 0;
                                                                dbginval();
                                                        }
                                                        break;
                                                }
                                                
                                                /* file by number - look up in line source list */
                                                for (i = 1, lin = ctx->dbgcxlin ; lin ;
                                                         lin = lin->linnxt, ++i)
                                                {
                                                        if (i == atoi(buf + 2)) break;
                                                }
                                                
                                                if (lin)
                                                {
                                                        lindef *oldlin;
                                                        
                                                        if (!dbgsrclin && dbgsrcfp)
                                                        {
                                                                fclose(dbgsrcfp);
                                                                dbgsrcfp = 0;
                                                        }
                                                        oldlin = dbgsrclin;
                                                        dbgswitch(&dbgsrclin, lin);
                                                        dbgsrcfp = ((linfdef *)lin)->linffp;
                                                        if (!dbgsrcfp)
                                                        {
                                                                dbgptf("can't display file - not found\n");
                                                                dbgswitch(&dbgsrclin, oldlin);
                                                        }
                                                        else
                                                        {
                                                                dbgsrctop = 0L;
                                                                osfseek(dbgsrcfp, 0L, OSFSK_SET);
                                                                dbginval();
                                                        }
                                                }
                                                else
                                                        dbgptf("invalid file number\n");
                                                break;
                                                
                                        default:
                                                dbgptf("invalid file command\n");
                                                break;
                                }
                                break;
                        
                        case '\0':
                                break;
                                
                        case 'q':
                                if (strcmp(buf, "quit"))
                                        dbgptf("invalid debugger command\n");
                                else
                                {
                                        /* make sure they really mean it, and quit if so */
                                        if (CautionAlert(12079, NULL) == 1)
                                        {
                                                exit(0);        /* quit immediately - already confirmed */
                                        }
                                }
                                break;
                                
                        case 'k':
                                dbgstktr(ctx, dbgudisp, NULL, level, false, false);
                                break;
                                
                        case 'h':
                                {
                                        extern int dbghid;
                                        
                                        if (!strcmp(buf, "help")) goto help_command;
                                        dbghid = !dbghid;
                                        dbgptf("Hidden text will be %s.\n", dbghid ? "displayed" : "suppressed");
                                }
                                break;
                                
                        case '/':
                                ++buf;          /* skip the '/' part */
                                if (*buf)
                                {
                                        strcpy(lastsearch, buf);
                                        lastpos = 1;
                                }
                                dbgusearch(dbgsrclin, dbgsrcfp, &lastpos, &dbgsrctop, lastsearch);
                                break;
                                                        
                        case 'u':
                                if (level < ctx->dbgcxfcn - 1) ++level;
                                dbgptf("evaluations are now at level %d\n", level+1);
                                break;
                                
                        case 'd':
                                if (level > 0) --level;
                                dbgptf("evaluations are now at level %d\n", level+1);
                                break;
        
                        case '?':
                        help_command:
                                {
                                        osfildef   *fp;
                                        #if 0
                                        /* Ideally, we fetch the help file from the application folder,
                                                but because we're not changing working directories, we can
                                                fake it by using a partial pathname. */
                                        char            buf[128];
                                        extern char     _argv[];
                                        
                                        strcpy(buf, _argv);
                                        strcat(buf, "HELP.TDB");
                                        #endif
                                        
                                        if (!(fp = osfoprt(/*buf*/ "help.tdb", OSFTTEXT)))
                                                dbgptf("HELP.TDB not found\n");
                                        else
                                        {
                                                if (!dbgsrclin && dbgsrcfp) fclose(dbgsrcfp);
                                                dbgswitch(&dbgsrclin, (lindef *)0);
                                                dbgsrcfp = fp;
                                                dbgsrctop = 0;
                                                dbginval();
                                        }
                                }
                                break;
                                
                        default:
                                dbgptf("invalid debugger command\n");
                                break;
                }
        }
}

void dbgucmd(dbgcxdef *ctx, int bphit, int err, unsigned int *exec_ofs)
{
        #pragma unused(exec_ofs)
        
        WindowPtr       oldWindow;
        int                     oldCmdpoint;
        extern int      cmdpoint;
        static int      dbgCmdpoint;
        
        os_flush();
        
        /* save current window information */
        oldWindow = myWindow;
        oldCmdpoint = cmdpoint;

        /* switch to debugger window */
        myWindow = dbgWindow;
        cmdpoint = dbgCmdpoint;

        /* turn on debugger menus */
        DisableItem(myMenus[3], 4);
        
        EnableItem(myMenus[2], 8);
        EnableItem(myMenus[2], 9);
        EnableItem(myMenus[3], 1);
        EnableItem(myMenus[3], 2);
        EnableItem(myMenus[3], 3);
        EnableItem(myMenus[3], 5);
        EnableItem(myMenus[3], 6);
        EnableItem(myMenus[3], 7);
        EnableItem(myMenus[4], 1);
        EnableItem(myMenus[4], 2);
        EnableItem(myMenus[4], 3);
        EnableItem(myMenus[4], 4);
        EnableItem(myMenus[5], 1);
        EnableItem(myMenus[5], 2);
        EnableItem(myMenus[5], 3);
        EnableItem(myMenus[5], 4);
        EnableItem(myMenus[5], 5);
        EnableItem(myMenus[6], 1);
        EnableItem(myMenus[6], 2);
        EnableItem(myMenus[6], 3);
        EnableItem(myMenus[7], 1);
        EnableItem(myMenus[7], 2);

        /* enter the main debugger command processor */
        if (bphit) dbgptf("Stop at breakpoint %d\n", bphit);
        else if (err) errprelog(ctx->dbgcxerr, err);
        dbgucmd1(ctx);

        /* turn off debugger menus */
        EnableItem(myMenus[3], 4);

        DisableItem(myMenus[2], 8);
        DisableItem(myMenus[2], 9);
        DisableItem(myMenus[3], 1);
        DisableItem(myMenus[3], 2);
        DisableItem(myMenus[3], 3);
        DisableItem(myMenus[3], 5);
        DisableItem(myMenus[3], 6);
        DisableItem(myMenus[3], 7);
        DisableItem(myMenus[4], 1);
        DisableItem(myMenus[4], 2);
        DisableItem(myMenus[4], 3);
        DisableItem(myMenus[4], 4);
        DisableItem(myMenus[5], 1);
        DisableItem(myMenus[5], 2);
        DisableItem(myMenus[5], 3);
        DisableItem(myMenus[5], 4);
        DisableItem(myMenus[5], 5);
        DisableItem(myMenus[6], 1);
        DisableItem(myMenus[6], 2);
        DisableItem(myMenus[6], 3);
        DisableItem(myMenus[7], 1);
        DisableItem(myMenus[7], 2);
                
        /* restore text window */
        myWindow = oldWindow;
        cmdpoint = oldCmdpoint;
}

void dbguini(dbgcxdef *ctx, const char *game_filename)
{
        #pragma unused(game_filename)
        
        dbgctx = ctx;
}

/* display an error message */
void dbguerr(dbgcxdef *ctx, int errno, char *msg)
{
        VARUSED(ctx);
        
        dbgptf("TADS-%d: %s\n", errno, msg);
}

/* output support for hidden output */
void trchid(void)
{
        outflushn(1);
        os_printf("---Hidden output---\n");
}

void trcsho(void)
{
        outflushn(1);
        os_printf("---End of hidden output, normal output resumes---\n");
}

static ListHandle       bp_list;
static DialogPtr        bp_dlg;
static int                      bp_type;
static int                      bp_cnt;

static void dbgbuildbp(void * /*ListHandle*/ mylist_v, char *str, int strl)
{
        ListHandle      mylist = (ListHandle) mylist_v;
        Cell            cell;
        static char     buf[128];
        static char *ptr = buf;
        static int      disabled;

        if (*str == '\n' || *str == '\r')
        {
                if (bp_type == 0                                                /* accept any bp */
                        || (bp_type == 1 && disabled)   /* accept only disabled */
                        || (bp_type == 2 && !disabled)) /* accept only enabled */
                {
                        if (mylist)
                        {
                                cell.v = LAddRow(1, 32767, mylist);
                                cell.h = 0;
                                LSetCell(buf, (int)(ptr - buf), cell, mylist);
                        }
                        else
                                ++bp_cnt;
                }
                ptr = buf;
                disabled = 0;
        }
        else if (strl == 11 && !memcmp(str, " [disabled]", (size_t)11))
                disabled = 1;
        else
        {
                memcpy(ptr, str, (size_t)strl);
                ptr += strl;
        }
}

static short list_btn_adjust;
static pascal Boolean dbgfilter(DialogPtr dlg, EventRecord *evt, short *item)
{
        Cell    cell;
        
        if (*item == 3) {
                Point where = evt->where;
                
                SetPort(GetDialogWindow(dlg));
                GlobalToLocal(&where);
                LClick(where, evt->modifiers, bp_list);
        }
        
        /* enable "Delete" button only if something's selected */
        cell.v = cell.h = 0;
        HiliteControlItem(bp_dlg, list_btn_adjust, LGetSelect(TRUE, &cell, bp_list) ? 0 : 255);
        
        if (evt->what == keyDown) {
                char key = evt->message & charCodeMask;
                if (key == 13 || key == 3) {
                        *item = 1;
                        return -1;
                }
        }
        
        return 0;
}

/* handle a breakpoint list box */
static void dbglist(int typ /* 0 = delete, 1 = enable, 2 = disable */)
{
        DialogPtr       mydlg;
        ListHandle      mylist;
        Rect            itemRect;
        Rect            dataBounds;
        Rect            r;
        Point           pt;
        short           itemHit;
        Cell            cell;
        GrafPtr         savePort;
        
        /* find out if there are any breakpoints first */
        bp_type = typ;
        list_btn_adjust = 2;
        bp_cnt = 0;
        dbgbplist(dbgctx, dbgbuildbp, (void *)0);
        if (!bp_cnt)
        {
                ParamText(typ == 0 ? "\pNo breakpoints set!"
                                                   : typ == 1 ? "\pNo disabled breakpoints!"
                                                                          : "\pNo enabled breakpoints!",
                                "\p", "\p", "\p");
                StopAlert(12370, NULL);
                return;
        }
        
        /* create the dialog */
        mydlg = GetNewDialog(23200, NULL, (WindowPtr) -1L);
        bp_dlg = mydlg;
        GetPort(&savePort);
        SetPort((GrafPtr)mydlg);

        /* change button 2's title if this is enable/disable */
        if (typ)
                SetControlItemTitle(mydlg, 2, typ == 1 ? "\pEnable" : "\pDisable");
        
        /* draw a framing rectangle around the list box */
        GetControlItemRect(mydlg, 3, &itemRect);
        r = itemRect;
        r.left--; r.top--;
        r.bottom++; r.right++;
        FrameRect(&r);
        
        /* create the list box itself */
        itemRect.right -= 15;   /* for scroll bar */
        dataBounds.top = dataBounds.left = 
                dataBounds.bottom = dataBounds.right = 0;
        pt.h = pt.v = 0;
        mylist = LNew(&itemRect, &dataBounds, pt, 0, (WindowPtr)mydlg, TRUE, FALSE, FALSE, TRUE);
        bp_list = mylist;
        LAddColumn(1, 0, mylist);
        
        /* populate the list box with breakpoints */
        dbgbplist(dbgctx, dbgbuildbp, mylist);
        dfltbtn(mydlg);

        /* run the dialog */
        do
        {
                /* enable "Delete" button only if something's selected */
                cell.v = cell.h = 0;
                HiliteControlItem(mydlg, 2, LGetSelect(TRUE, &cell, mylist) ? 0 : 255);
                
                {
                        ModalFilterUPP  dbgfilterUPP;
                        
                        dbgfilterUPP = NewModalFilterProc(dbgfilter);
                        if (dbgfilterUPP) {
                                ModalDialog(dbgfilterUPP, &itemHit);
                                DisposeRoutineDescriptor(dbgfilterUPP);
                        }
                        else
                                itemHit = 0;
                }
                
                if (itemHit == 2)
                {
                        cell.v = cell.h = 0;
                        while (LGetSelect(TRUE, &cell, mylist))
                        {
                                char    buf[128];
                                short   len = 128;
                                int             id;
                                
                                LGetCell(buf, &len, cell, mylist);
                                id = atoi(buf);
                                switch(typ)
                                {
                                case 0:
                                        dbgbpdel(dbgctx, id);
                                        dbgugetbp(dbgctx);
                                        break;
                                        
                                case 1:
                                case 2:
                                        dbgbpdis(dbgctx, id, typ == 2);
                                        break;
                                }
                                LDelRow(1, cell.v, mylist);
                        }
                }
        } while (itemHit != 1);
        
        SetPort(savePort);
        DisposeDialog(mydlg);
}

/* select a source module */
static void dbgselmod(void)
{
        DialogPtr       mydlg;
        ListHandle      mylist;
        Rect            itemRect;
        Rect            dataBounds;
        Rect            r;
        Point           pt;
        short           itemHit;
        Cell            cell;
        GrafPtr         savePort;
        lindef     *lin;
        int                     cnt;

        /* find out if there are any modules first */
        for (cnt = 0, lin = dbgctx->dbgcxlin ; lin ; lin = lin->linnxt)
        {
                /* if (((linfdef *)lin)->linffp) */
                ++cnt;
        }
        if (!cnt)
        {
                ParamText("\pNo modules available!", "\p", "\p", "\p");
                StopAlert(12370, NULL);
                return;
        }
        
        /* create the dialog */
        mydlg = GetNewDialog(23200, NULL, (WindowPtr) -1L);
        bp_dlg = mydlg;
        list_btn_adjust = 1;
        GetPort(&savePort);
        SetPort((GrafPtr)mydlg);

        /* change button titles to "OK" and "Cancel" */
        SetControlItemTitle(mydlg, 1, "\pOK");
        SetControlItemTitle(mydlg, 2, "\pCancel");
        
        /* draw a framing rectangle around the list box */
        GetControlItemRect(mydlg, 3, &itemRect);
        r = itemRect;
        r.left--; r.top--;
        r.bottom++; r.right++;
        FrameRect(&r);
                
        /* create the list box itself */
        itemRect.right -= 15;   /* for scroll bar */
        dataBounds.top = dataBounds.left = dataBounds.bottom = dataBounds.right = 0;
        pt.h = pt.v = 0;
        mylist = LNew(&itemRect, &dataBounds, pt, 0, (WindowPtr)mydlg, TRUE, FALSE, FALSE, TRUE);
        (*mylist)->selFlags = -128;
        bp_list = mylist;
        LAddColumn(1, 0, mylist);
        
        /* populate the list box with modules */
        for (lin = dbgctx->dbgcxlin ; lin ; lin = lin->linnxt)
        {
                /*if (((linfdef *)lin)->linffp)*/
                {
                        char *nm = ((linfdef *)lin)->linfnam;
                        
                        cell.v = LAddRow(1, 32767, mylist);
                        cell.h = 0;
                        LSetCell(nm, (int)strlen(nm), cell, mylist);
                }
        }
        
        /* run the dialog */
        dfltbtn(mydlg);
        do
        {
                /* enable "OK" button only if something's selected */
                cell.v = cell.h = 0;
                HiliteControlItem(mydlg, 1, LGetSelect(TRUE, &cell, mylist) ? 0 : 255);
                
                {
                        ModalFilterUPP  dbgfilterUPP;
                        
                        dbgfilterUPP = NewModalFilterProc(dbgfilter);
                        if (dbgfilterUPP) {
                                ModalDialog(dbgfilterUPP, &itemHit);
                                DisposeRoutineDescriptor(dbgfilterUPP);
                        }
                        else
                                itemHit = 0;
                }
                
                if (itemHit == 1)
                {
                        char    buf[128];
                        short   len = (short) sizeof(buf);
                        
                        cell.v = cell.h = 0;
                        LGetSelect(TRUE, &cell, mylist);
                        LGetCell(buf, &len, cell, mylist);
                        buf[len] = '\0';
                        
                        /* figure out which one it is */
                        for (lin = dbgctx->dbgcxlin ; lin ; lin = lin->linnxt)
                        {
                                /*if (((linfdef *)lin)->linffp)*/
                                {
                                        char *nm = ((linfdef *)lin)->linfnam;
                                        if (!strcmp(nm, buf))
                                        {
                                                if (!dbgsrclin && dbgsrcfp) fclose(dbgsrcfp);
                                                dbgswitch(&dbgsrclin, lin);
                                                dbgsrcfp = ((linfdef *)lin)->linffp;
                                                dbgsrctop = 0;
                                                dbginval();
                                                goto done;
                                        }
                                }
                        }
                }
        } while (itemHit != 1 && itemHit != 2);

done:
        SetPort(savePort);
        DisposeDialog(mydlg);
}

/* run a menu dialog */
static char *dbg_dlg(short id, char *buf, char *cmd, char *str2)
{
        short           itemHit;
        DialogPtr       mydlg;
        char       *ret = NULL;
        Str255          textval;
        
        mydlg = GetNewDialog(id, NULL, (WindowPtr) -1L);
        dfltbtn(mydlg);
        
        do
        {
                ModalDialog(NULL, &itemHit);
        } while (itemHit != 1 && itemHit != 2);
        if (itemHit == 1)
        {
                ret = buf;
                strcpy(buf, cmd);
                buf += strlen(buf);
                
                GetControlItemTitle(mydlg, 4, textval);
                StringCat_cp(buf, textval, 0);
                buf += StrLength(textval);
                
                if (str2)
                {
                        GetControlItemTitle(mydlg, 6, textval);
                        if (StrLength(textval) != 0)
                        {
                                strcpy(buf, str2);
                                buf += strlen(buf);
                                StringCat_cp(buf, textval, 0);
                                buf += StrLength(textval);
                        }
                }
                *buf = '\0';
        }
        DisposeDialog(mydlg);
        return(ret);
}

char *dbgmenu(long mResult)
{
        short           theItem;
        static char     newbuf[255];
        char       *newcmd = (char *)0;
        
        theItem = LoWord(mResult);
        switch (HiWord(mResult))
        {
                case 129: /* file */
                        switch(theItem)
                        {
                                case 1:
                                        newcmd = "quit";
                                        break;
                        }
                        break;
                        
                case 130: /* edit */
                        switch(theItem)
                        {
                                case 8: /* find... */
                                        newcmd = dbg_dlg(14306, newbuf, "/", (char *)0);
                                        break;
                                        
                                case 9: /* find again */
                                        newcmd = "/";
                                        break;
                        }
                        break;
                        
                case 131: /* execution */
                        switch(theItem)
                        {
                                case 1: /* go */
                                        newcmd = "g";
                                        break;
                                        
                                case 2: /* step */
                                        newcmd = "t";
                                        break;
                                        
                                case 3: /* step over */
                                        newcmd = "p";
                                        break;
                                        
                                case 4: /* Stop */
                                        newcmd = " ";
                                        dbgctx->dbgcxflg |= DBGCXFSS;
                                        dbgctx->dbgcxflg &= ~DBGCXFSO;
                                        break;
                
                                case 5: /* call trace on/off */
                                        if (dbgctx->dbgcxflg & DBGCXFTRC)
                                                dbgctx->dbgcxflg &= ~DBGCXFTRC;
                                        else
                                                dbgctx->dbgcxflg |= DBGCXFTRC;
                                        SetItemMark(myMenus[3], 5,
                                                                (dbgctx->dbgcxflg & DBGCXFTRC) ? 0x12 : 0x00);
                                        break;
                
                                case 6: /* show call history */
                                        newcmd = "c";
                                        break;
                
                                case 7: /* clear call history */
                                        newcmd = "cc";
                                        break;
                        }
                        break;
                        
                case 132: /* data */
                        switch(theItem)
                        {
                                case 1: /* examine */
                                        newcmd = dbg_dlg(25121, newbuf, "e ", (char *)0);
                                        break;
                                        
                                case 2: /* stack */
                                        newcmd = "k";
                                        break;
                                        
                                case 3: /* up */
                                        newcmd = "u";
                                        break;
                                        
                                case 4: /* down */
                                        newcmd = "d";
                                        break;
                        }
                        break;
                        
                case 133: /* breakpoint */
                        switch(theItem)
                        {
                                case 1: /* set */
                                        newcmd = dbg_dlg(26375, newbuf, "bp ", " when ");
                                        break;
                                        
                                case 2: /* clear */
                                        dbglist(0);
                                        break;
                                        
                                case 3: /* disable */
                                        dbglist(2);
                                        break;
                                        
                                case 4: /* enable */
                                        dbglist(1);
                                        break;
                                        
                                case 5: /* list */
                                        newcmd = "bl";
                                        break;
                        }
                        break;
                        
                case 134: /* source */
                        switch(theItem)
                        {
                                case 1: /* view module */
                                        dbgselmod();
                                        break;
                                        
                                case 2: /* view file */
                                        {
                                                Str255  fname;
                                                
                                                if (!os0askfile("View File", fname, 'TEXT', 0))
                                                {
                                                        char            buf[256];
                                                        osfildef   *fp;
                                                        
                                                        StringCopy_cp(buf, fname, sizeof(buf));
                                                        if ((fp = osfoprt(buf, OSFTTEXT)) != NULL)
                                                        {
                                                                if (!dbgsrclin && dbgsrcfp) fclose(dbgsrcfp);
                                                                dbgswitch(&dbgsrclin, (lindef *)0);
                                                                dbgsrcfp = fp;
                                                                dbgsrctop = 0;
                                                                dbginval();
                                                        }
                                                        else
                                                        {
                                                                ParamText("\pCan't open file!", "\p", "\p", "\p");
                                                                StopAlert(12370, NULL);
                                                        }
                                                }
                                        }
                                        break;
                                
                                case 3: /* home */
                                        if (dbgsrclinhome
                                                && (dbgsrctop != dbgsrchome || dbgsrclin != dbgsrclinhome))
                                        {
                                                if (!dbgsrclin && dbgsrcfp) fclose(dbgsrcfp);
                                                dbgswitch(&dbgsrclin, dbgsrclinhome);
                                                dbgsrcfp = ((linfdef *)dbgsrclin)->linffp;
                                                osfseek(dbgsrcfp, dbgsrchome, OSFSK_SET);
                                                dbgsrctop = dbgsrchome;
                                                dbginval();
                                        }
                                        break;
                        }
                        break;
                
                case 135: /* watch */
                        switch(theItem)
                        {
                                case 1: /* set... */
                                        newcmd = dbg_dlg(14843, newbuf, "ws ", (char *)0);
                                        break;
                                        
                                case 2: /* delete... */
                                        break;
                        }
                        break;
        }
        
        return(newcmd);
}
