/*
==========================================================================================

	File:		mdebug.c
	
	Contains:	Debugging routines for Mac OS.

	Copyright:	 1998 by Christopher Nebel, all rights reserved.

	Change History (most recent first):

		 <2>	 8/29/98	Neb		Only include printmem if DEBUG_IO is on.
		 <1>	 7/23/98	Neb		Spliced off from osmacsup.c.

=========================================================================================
*/

#ifndef NDEBUG

#include "mdebug.h"

#include <ctype.h>
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>  // Included so we can detect the version.
#include <Types.h>

#include "mstring.h"


int dprintf(const char *fmt, ...)
{
	Str255	buf;
	va_list	argptr;
	int		ret;
	
	va_start(argptr, fmt);
	ret = pvsprintf(buf, fmt, argptr);
	va_end(argptr);
	
	DebugStr(buf);
	return ret;
}


#if DEBUG_IO

void printmem(const void *pv, size_t len)
{
	size_t		i;
	const char *p;
	
	for (i = 0, p = (const char *) pv; i < len; i++, p++) {
		if (*p == '\t')
			printf("\\t");
		else if (*p == '\n')
			printf("\\n");
		else if (*p == '\r')
			printf("\\r");
		else if (iscntrl(*p))
			printf(".");
		else
			printf("%c", *p);
	}
}

#endif


/*
	Define routines so assert writes its messages using dprintf.

	The MPW libraries define an __assertprint routine;
	the MSL libraries define an __assertion_failed routine.
	
	To be compatible with either, override them both.
*/

#ifdef __cassert__
static /* Blocks a "no prototype" warning. */
#endif
void __assertprint(const char* file, int line, const char* expr)
{
	dprintf("File \"%s\", Line %d # Assertion %s failed!\n", file, line, expr);
}

#ifdef __cassert__
/* Using the MSL headers. */
void __assertion_failed(char * expr, char * file, int line)
{
	__assertprint(file, line, expr);
}
#endif

#endif
