#ifdef RCSID
static char RCSid[] =
"$Header: d:/cvsroot/tads/TADS2/msdos/OSDOSEX.C,v 1.2 1999/05/17 02:52:18 MJRoberts Exp $";
#endif

/* Copyright (c) 1997 by Michael J. Roberts.  All Rights Reserved. */
/*
Name
  osdosex.c - external function implementation for DOS
Function
  
Notes
  
Modified
  11/02/97 MJRoberts  - Creation
*/

#ifndef DJGPP
#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <share.h>
#endif /* DJGPP */

#include "os.h"

#if defined(__DPMI16__) || defined(__WIN32__)
# include <windows.h>
#endif

#ifdef DJGPP

/*
 *   os_exeseek - opens the given .EXE file and seeks to the end of the
 *   executable part of it, on the presumption that a datafile is to be
 *   found there.  
 */
osfildef *os_exeseek(const char *exefile, char *typ)
{
    return((osfildef *)0);
}

/*
 *   os_exfld - load in an external function from an open file, given the
 *   size of the function (in bytes).  Returns a pointer to the newly
 *   allocated memory block containing the function in memory.  
 */
int (*os_exfld( osfildef *fp, unsigned len ))(void *)
{
    return  (int (*)(void *)) 0;
}

/*
 *   Load an external function from a file.  This routine assumes that the
 *   file has the same name as the resource.  
 */
int (*os_exfil(const char *name))(void *)
{
    return (int (*)(void *)) 0;
}

/*
 *   call an external function, passing it an argument (a string pointer),
 *   and passing back the string pointer returned by the external function 
 */
int os_excall(int (*extfn)(void *), void *arg)
{
    return 0;
}

#else /* DJGPP */

/*
 *   os_exeseek - opens the given .EXE file and seeks to the end of the
 *   executable part of it, on the presumption that a datafile is to be
 *   found there.  
 */
osfildef *os_exeseek(const char *exefile, char *typ)
{
    osfildef *fp;
    unsigned  ofs, segcnt;
    unsigned  long seekpt;
    unsigned  long check;
    unsigned  long startofs;
    unsigned  long end;
    char      typbuf[4];

    /* open the file and seek to the very end */
    if (!(fp = _fsopen(exefile, "rb", SH_DENYWR))) return((FILE *)0);
    fseek(fp, 0L, SEEK_END);

    /* look through tagged blocks until we find the type we want */
    for (;;)
    {
        /* seek back past the descriptor block */
        fseek(fp, -12L, SEEK_CUR);
        seekpt = ftell(fp);

        /* read the elements of the descriptor block */
        if (fread(&check, sizeof(check), 1, fp) != 1
            || fread(typbuf, sizeof(typbuf), 1, fp) != 1
            || fread(&startofs, sizeof(startofs), 1, fp) != 1)
            break;

        /* check the signature to make sure we're looking at a valid block */
        if (check != ~seekpt)
            break;

        /* seek to the start of the data for this resource */
        fseek(fp, startofs, SEEK_SET);

        /* if this is the one we want, return it, otherwise keep looking */
        if (!memcmp(typ, typbuf, sizeof(typbuf)))
        {
            /* check the header to make sure it matches */
            if (fread(&check, sizeof(check), 1, fp) != 1
                || fread(typbuf, sizeof(typbuf), 1, fp) != 1
                || check != ~startofs
                || memcmp(typ, typbuf, sizeof(typbuf)))
                break;
            return fp;
        }
    }

    /*
     *   we didn't find it - if the caller is asking for the game file
     *   segment ("TGAM"), try the old way, in case this got built with
     *   the old maketrx program 
     */
    fseek(fp, 0L, SEEK_SET);
    if (fread(&ofs, sizeof(ofs), 1, fp) != 1 ||
        fread(&ofs, sizeof(ofs), 1, fp) != 1 ||
        fread(&segcnt, sizeof(segcnt), 1, fp) != 1)
        goto no_data;

    seekpt = ((unsigned long)segcnt - 1)*512L + (unsigned long)ofs;
    if (fseek(fp, seekpt, SEEK_SET)) goto no_data;

    /* check for proper signature, to make sure it's not the symbol table */
    if (fread(&end, sizeof(end), 1, fp) != 1
        || fseek(fp, 0L, SEEK_END)
        || (unsigned long)ftell(fp) != end
        || fseek(fp, seekpt + sizeof(end), SEEK_SET))
        goto no_data;

    /* success! */
    return(fp);

no_data:

    /* we didn't find anything - close the file and return failure */
    fclose(fp);
    return((FILE *)0);
}

/* ------------------------------------------------------------------------ */
/*
 *   Real-mode and 16-bit protected mode external function interfaces 
 */

#ifndef __WIN32__

static void *canon(void *ptr)
{
    unsigned long p = (unsigned long)ptr;
    unsigned long abs = ((p >> 16) << 4) + (p & 0xffff);
    if (abs & (unsigned long)0xf)
        abs = (abs & ~(unsigned long)0xf) + (unsigned long)0x10;
    return((void *)(abs << 12));
}

/*
 *   os_exfld - load in an external function from an open file, given the
 *   size of the function (in bytes).  Returns a pointer to the newly
 *   allocated memory block containing the function in memory.  
 */
int (*os_exfld(osfildef *fp, unsigned len))(void *)
{
    void      *extfn;
    unsigned   alo;

#ifdef MSOS2
    /* for OS/2, don't load anything, but seek past the resource */
# ifdef MICROSOFT
    if (_osmode == OS2_MODE)
# endif /* MICROSOFT */
    {
        osfseek(fp, (long)len, OSFSK_SET);
        return((int (*)(void))0);
    }
#endif /* MSOS2 */

#ifdef __DPMI16__

    HANDLE selector;

    selector = GlobalAlloc(GMEM_FIXED|GMEM_ZEROINIT, len);
    if (!selector || !(extfn = GlobalLock(selector)))
        return 0;

    /* read the file */
    osfrb(fp, extfn, len);

    /* change the selector to a code segment */
    _asm
    {
        mov   bx, selector                                  /* get selector */
            lar   ax, bx                       /* get current access rights */
            mov   cl, ah
            or    cl, 8               /* set the CODE bit in the descriptor */
            mov   ax, 9                 /* function = set descriptor rights */
            mov   ch, 0
            int   31h
    }

    /* close the file and return the pointer to the function in memory */
    return((int (*)(void *))extfn);

#else /* __DPMI16 __ */

    /* figure out how much memory is needed and allocate it */
    alo = ((len + 0xf) & ~0xf) + 16;    /* round to mult of 16, plus 1 page */
    extfn = canon(malloc(alo));/* allocate the memory, canonicalize pointer */
    if (!extfn)
        return((int (*)(void *))0 );

    /* read the file */
    osfrb(fp, extfn, len);

    /* close the file and return the pointer to the function in memory */
    return((int (*)(void *))extfn);

#endif /* __DPMI16__ */
}

/*
 *   Load an external function from a file.  This routine assumes that the
 *   file has the same name as the resource. 
 */
int (*os_exfil(const char *name))(void *)
{
    FILE      *fp;
    unsigned   len;
    int      (*extfn)(void *);

#ifdef MSOS2
    /* load the function from a DLL of the same name as the function */
# ifdef MICROSOFT
    if (_osmode == OS2_MODE)
# endif /* MICROSOFT */
    {
        CHAR    failname[128];
        HMODULE hmod;
        PFN     pfn;

        if (DosLoadModule(failname, sizeof(failname), (PSZ)name, &hmod)
            || DosGetProcAddr(hmod, (PSZ)"_main", &pfn))
        {
            return((int (*)(void))0);
        }
        return((int (*)(void))pfn);
    }
#endif /* MSOS2 */

#ifdef __DPMI16__

#endif

    /* open the file and see how big it is to determine our memory needs */
    if ( !( fp = fopen( name, "rb" ))) return( (int (*)(void *))0 );
    (void)fseek( fp, 0L, 2 );
    len = (unsigned)ftell( fp );                    /* total length of file */

    (void)fseek( fp, 0L, 0 );
    extfn = os_exfld(fp, len);

    fclose( fp );
    return( extfn );
}

/*
 *   call an external function, passing it an argument (a string pointer),
 *   and passing back the string pointer returned by the external function
 */
int os_excall(int (*extfn)(void *), void *arg)
{
    return((*extfn)(arg));
}

#endif /* !__WIN32__ */

/* ------------------------------------------------------------------------ */
/*
 *   Win32 external function interfaces - not implemented at present 
 */

#ifdef __WIN32__

int (*os_exfld(osfildef *fp, unsigned len))(void *)
{
    /* NOT IMPLEMENTED - scan past the resource and fail */
    osfseek(fp, (long)len, OSFSK_CUR);
    return 0;
}

int (*os_exfil(const char *name))(void *)
{
    HINSTANCE hlib;
    FARPROC proc;
    
    /* 
     *   load the library of the given name; if we can't load the library,
     *   we can't load the external function 
     */
    hlib = LoadLibrary(name);
    if (hlib == 0)
        return 0;

    /* get the address of the "_main" procedure */
    proc = GetProcAddress(hlib, "main");

    /* if that failed, unload the library and return failure */
    if (proc == 0)
    {
        FreeLibrary(hlib);
        return 0;
    }

    /* 
     *   return the procedure address, suitably cast; unfortunately, we
     *   have no provision for freeing the library instance handle, so
     *   we'll just have to count on Windows releasing it when the process
     *   terminates 
     */
    return (int (*)(void *))proc;
}

int os_excall(int (*extfn)(void *), void *arg)
{
    /* call the function directly */
    return((*extfn)(arg));
}

#endif /* __WIN32__ */
#endif /* DJGPP */

